@@ -4,11 +4,12 @@ use crate::instrumentation_profile::types::*;
 use crate::util::*;
 use anyhow::{bail, Result};
 use nom::error::Error as NomError;
-use object::{Endian, Endianness, Object, ObjectSection, Section};
+use object::{Endian, Endianness, Object, ObjectSection, ReadCache, ReadRef, Section};
 use std::convert::TryInto;
 use std::error::Error;
 use std::fmt;
 use std::fs;
+use std::io::BufReader;
 use std::path::{Path, PathBuf};
 use tracing::{debug, error, trace, warn};
 
@@ -61,8 +62,22 @@ impl Error for SectionReadError {}
 pub fn read_object_file(object: &Path, version: u64) -> Result<CoverageMappingInfo> {
     // I believe vnode sections added by llvm are unnecessary
 
-    let binary_data = fs::read(object)?;
-    let object_file = object::File::parse(&*binary_data)?;
+    let binary_data = ReadCache::new(BufReader::new(fs::File::open(object)?));
+    let object_file = object::File::parse(&binary_data)?;
+
+    let prof_counts = object_file
+        .section_by_name("__llvm_prf_cnts")
+        .or(object_file.section_by_name(".lprfc"))
+        .and_then(|x| parse_profile_counters(object_file.endianness(), &x).ok());
+
+    debug!("Parsed prf_cnts: {:?}", prof_counts);
+
+    let prof_data = object_file
+        .section_by_name("__llvm_prf_data")
+        .or(object_file.section_by_name(".lprfd"))
+        .and_then(|x| parse_profile_data(object_file.endianness(), &x).ok());
+
+    debug!("Parsed prf_data section: {:?}", prof_data);
 
     let cov_fun = object_file
         .section_by_name("__llvm_covfun")
@@ -82,20 +97,6 @@ pub fn read_object_file(object: &Path, version: u64) -> Result<CoverageMappingIn
 
     debug!("Parsed covmap section: {:?}", cov_map);
 
-    let prof_counts = object_file
-        .section_by_name("__llvm_prf_cnts")
-        .or(object_file.section_by_name(".lprfc"))
-        .and_then(|x| parse_profile_counters(object_file.endianness(), &x).ok());
-
-    debug!("Parsed prf_cnts: {:?}", prof_counts);
-
-    let prof_data = object_file
-        .section_by_name("__llvm_prf_data")
-        .or(object_file.section_by_name(".lprfd"))
-        .and_then(|x| parse_profile_data(object_file.endianness(), &x).ok());
-
-    debug!("Parsed prf_data section: {:?}", prof_data);
-
     Ok(CoverageMappingInfo {
         cov_map,
         cov_fun,
@@ -140,6 +141,7 @@ impl<'a> CoverageMapping<'a> {
         result.insert(Counter::default(), 0);
         let record = self.profile.find_record_by_hash(func.header.name_hash);
         if let Some(func_record) = record.as_ref() {
+            result.reserve(func_record.record.counts.len());
             for (id, count) in func_record.record.counts.iter().enumerate() {
                 result.insert(Counter::instrumentation(id as u64), *count as i64);
             }
@@ -174,7 +176,7 @@ impl<'a> CoverageMapping<'a> {
 
                 let mut pending_exprs = vec![];
 
-                for (expr_index, expr) in func.expressions.iter().enumerate() {
+                for (expr_index, expr) in func.expressions.iter().enumerate().rev() {
                     let lhs = region_ids.get(&expr.lhs);
                     let rhs = region_ids.get(&expr.rhs);
                     match (lhs, rhs) {
@@ -182,11 +184,13 @@ impl<'a> CoverageMapping<'a> {
                             let count: i64 = match expr.kind {
                                 ExprKind::Subtract => {
                                     trace!("Subtracting counts: {} - {}", lhs, rhs);
-                                    lhs - rhs
+                                    // TODO these saturating ops need investigation. Potentially
+                                    // something is wrong.
+                                    lhs.saturating_sub(*rhs)
                                 }
                                 ExprKind::Add => {
                                     trace!("Adding counts: {} + {}", lhs, rhs);
-                                    lhs + rhs
+                                    lhs.saturating_add(*rhs)
                                 }
                             };
 
@@ -280,9 +284,9 @@ impl<'a> CoverageMapping<'a> {
     }
 }
 
-fn parse_coverage_mapping(
+fn parse_coverage_mapping<'data, R: ReadRef<'data>>(
     endian: Endianness,
-    section: &Section<'_, '_>,
+    section: &Section<'data, '_, R>,
     version: u64,
 ) -> Result<FxHashMap<u64, Vec<PathBuf>>, SectionReadError> {
     if let Ok(mut data) = section.data() {
@@ -322,9 +326,9 @@ fn parse_coverage_mapping(
     }
 }
 
-fn parse_coverage_functions(
+fn parse_coverage_functions<'data, R: ReadRef<'data>>(
     endian: Endianness,
-    section: &Section<'_, '_>,
+    section: &Section<'data, '_, R>,
 ) -> Result<Vec<FunctionRecordV3>, SectionReadError> {
     debug!("Parsing coverage functions");
     if let Ok(original_data) = section.data() {
@@ -485,9 +489,9 @@ fn parse_mapping_regions<'a>(
     Ok((bytes, mapping))
 }
 
-fn parse_profile_data(
+fn parse_profile_data<'data, R: ReadRef<'data>>(
     endian: Endianness,
-    section: &Section<'_, '_>,
+    section: &Section<'data, '_, R>,
 ) -> Result<Vec<ProfileData>, SectionReadError> {
     if let Ok(data) = section.data() {
         let mut bytes = data;
@@ -528,9 +532,9 @@ fn parse_profile_data(
     }
 }
 
-fn parse_profile_counters(
+fn parse_profile_counters<'data, R: ReadRef<'data>>(
     endian: Endianness,
-    section: &Section<'_, '_>,
+    section: &Section<'data, '_, R>,
 ) -> Result<Vec<u64>, SectionReadError> {
     if let Ok(data) = section.data() {
         let mut result = vec![];
